#ifndef DATASET_H
#define DATASET_H

#include <vector>
#include <string>
#include <set>
#include <utility>
#include <map>
#include <algorithm>

#include "args.hpp"
#include "misc.hpp"

using namespace std;

class DataSet
{

public:
  
  friend ostream & operator<<(ostream &, const DataSet &);
  
  DataSet(string, int, long);
  ~DataSet();

  inline long size() { return _data.size(); }
  inline int numCategories() { return _categories.size(); }
  inline int dim() { return _dim; }
  inline long getCategoryCount(int c) { return _categoryCount[_categoryMap[c]]; }
  
  int getCategory(int); // Get category at some index.
  int getCategoryIndex(int); // Get index of some category.
  
  double * feature(long); // Get feature vector.
  int category(long); // Get class of pattern.

  string _dataFile; // Name of data file.

  // _categories stores the categories sorted into increasing order. _categoryCount[i]
  // stores the number of patterns of class _categories[i]. _categoryMap[c] stores
  // the index in _categories in which category c is stored.

  int _dim; // Number of features.
  vector < int > _categories; // Categories, sorted by id in increasing order.
  vector < long > _categoryCount; // Number of patterns of each category.
  map < int, int > _categoryMap; // Map category->index (index in _categories).
										     
  vector < pair < double *, int > * > _data; // The patterns themselves.

};

#endif
