#!/usr/bin/perl

use strict;

my $STARTSEED = 1;

my $NUMFEATURES = 21;
my $TRAINDATAFILE = "ann-train.data";
my $TRAINNUMPATTERNS = 3772;
my $TESTDATAFILE = "ann-test.data";
my $TESTNUMPATTERNS = 3428;

my $NUMAGENTS = 25;
my $NUMTESTS = 25;
my $NUMEPOCHS = 50000;
#my $PASIZE = ;
#my $PPSIZE = 25;

my $MININDSIZE = 1;
my $MAXINDSIZE = 48;

my $PBIDMUTATE = 0.5;
my $PBIDSWAP = 0.5;
my $PBIDDELETE = 0.5;
my $PBIDADD = 0.5;
my $PACTIONMUTATE = 0.5;

# my $OUTCOMEALGORITHM = "SWB";
# my $OUTCOMEALGORITHM = "MWT";

my $OUTPUTMOD = 5000;

if(scalar(@ARGV) != 2)
{
    die "usage: genScript.pl prefix numRuns";
}

my $prefix = $ARGV[0];
my $numRuns = $ARGV[1];

print "prefix: $prefix\n";
print "numRuns: $numRuns\n";

my $run;
my $argFile;
my $seed;

open(RUN, ">$prefix" . ".run") || die "cannot open RUN $prefix.run";

for($run = 0; $run < $numRuns; $run++)
{
    $seed = $STARTSEED + $run;

    $argFile = "$prefix.$seed.arg";
    
    open(ARG, ">$argFile") || die "cannot open $argFile";

    print ARG "long seed $seed\n";
    print ARG "\n";
    
    print ARG "string prefix $prefix.$seed\n";
    print ARG "\n";

    print ARG "long numFeatures $NUMFEATURES\n";
    print ARG "string trainDataFile $TRAINDATAFILE\n";
    print ARG "long trainNumPatterns $TRAINNUMPATTERNS\n";
    print ARG "string testDataFile $TESTDATAFILE\n";
    print ARG "long testNumPatterns $TESTNUMPATTERNS\n";
    print ARG "\n";

    print ARG "long numAgents $NUMAGENTS\n";
    print ARG "long numTests $NUMTESTS\n";
    print ARG "long numEpochs $NUMEPOCHS\n";
#    print ARG "long paSize $PASIZE\n";
#    print ARG "long ppSize $PPSIZE\n";
    print ARG "\n";

    print ARG "long minIndSize $MININDSIZE\n";
    print ARG "long maxIndSize $MAXINDSIZE\n";

    print ARG "double pBidMutate $PBIDMUTATE\n";
    print ARG "double pBidSwap $PBIDSWAP\n";
    print ARG "double pBidDelete $PBIDDELETE\n";
    print ARG "double pBidAdd $PBIDADD\n";
    print ARG "double pActionMutate $PACTIONMUTATE\n";
    print ARG "\n";

#    print ARG "string outcomeAlgorithm $OUTCOMEALGORITHM\n";
#    print ARG "\n";

    print ARG "long outputMod $OUTPUTMOD\n";
    
    close(ARG);
    
    print RUN "./ccm $argFile 1> $prefix.$seed.std 2> $prefix.$seed.err\n";
}

close(RUN);

# # The variables in UPPERCASE can be changed by the user.

# my $COMMAND = "java EvoD"; # The EvoD program file name minus the arguments, e.g., "java EvoD" or "./a.out".

# my $RUNS = 50; # Number of different initializations per configuration, should be at least 30.

# my $POPSIZE = 250; # Population size.
# my $MININDSIZE = 1; # Minimum individual size.
# my $MAXINDSIZE = 40; # Maximum individual size.
# my $MAXITER = 10000; # Number of iterations (tournaments).
# my $PX = 0.9; # Probability of crossover.
# my $PM = 0.5; # Probability of mutation.
# my $PS = 0.5; # Probability of swap.
# my $TRAINDATA = "breast.data.train"; # Training data.
# my $TESTDATA = "breast.data.test"; # Test data.
# my $RULEPOOL = "Rules_Pool.txt"; # Pool of rules.

# my $OUTFILE = "EvoD.results"; # Where to write summary of results.
# # Each line will contain the identifier and seed for the run as well as the train hits, test hits, 
# # and size corresponding to the best individual in the run (in that order).

# my $run;
# my $prefix;
# my $indSize;

# my $line;
# my @toks;

# open(OUTFILE, ">$OUTFILE");

# # Perform the variable-size runs.

# for($run = 0; $run < $RUNS; $run++)
# {
#     $prefix = "var_$run";
#     print "$COMMAND $prefix $POPSIZE $MININDSIZE $MAXINDSIZE $MAXITER $PX $PM $PS $TRAINDATA $TESTDATA $RULEPOOL $run\n";
#     print STDERR `$COMMAND $prefix $POPSIZE $MININDSIZE $MAXINDSIZE $MAXITER $PX $PM $PS $TRAINDATA $TESTDATA $RULEPOOL $run`;

#     open(INFILE, "<$prefix.best");
#     $line = <INFILE>;
#     print OUTFILE "$prefix $run $line";
#     close(INFILE);
# }

# # Perform the fixed-size runs.

# for($indSize = $MININDSIZE; $indSize <= $MAXINDSIZE; $indSize++)
# {
#     for($run = 0; $run < $RUNS; $run++)
#     {
# 	$prefix = "fix_" . $indSize . "_$run";
# 	print "$COMMAND $prefix $POPSIZE $indSize $indSize $MAXITER $PX $PM $PS $TRAINDATA $TESTDATA $RULEPOOL $run\n";
# 	print STDERR `$COMMAND $prefix $POPSIZE $indSize $indSize $MAXITER $PX $PM $PS $TRAINDATA $TESTDATA $RULEPOOL $run`;

# 	open(INFILE, "<$prefix.best");
# 	$line = <INFILE>;
# 	print OUTFILE "$prefix $run $line";
# 	close(INFILE);
#     }
# }

# close(OUTFILE);


