#include "stats.hpp"


void Stats::update(Pop &pop, long epoch)
{
  vector < Agent * >::iterator iter;
  vector < Agent * > activeAgents;

  vector < double > maxBidMatrix; // stores max bids for the current pattern only

  int matInd;

  long curr = numRecords(); // current index into stats vectors

  _confMatrix.clear();
  _hitMap.clear();
  _missMap.clear();
  _actionCount.clear();
  _sumBidMatrix.clear();
  _sumMaxBidMatrix.clear();

  for(int i = 0; i < _data->_categories.size() * _data->_categories.size(); i++)
    {
      _confMatrix.push_back(0);
      _sumBidMatrix.push_back(0);
      _sumMaxBidMatrix.push_back(0);

      maxBidMatrix.push_back(0);
    }
  
  for(int i = 0; i < _data->_categories.size(); i++)
    {
      _actionCount.push_back(0);
    }

  _epochs.push_back(epoch);

  _meanAge.push_back(0);
  _meanSize.push_back(0);
  _meanRate.push_back(0);
  _meanProfit.push_back(0);
  _numActive.push_back(0);

  _numWinners.push_back(0);

  _hits.push_back(0);

  // calculate stats for active agents

  for(iter = pop._agents.begin(); iter != pop._agents.end(); iter++)
    {
      if((*iter)->_active == true) 
	{
	  //	  _hitMap.insert(map <long, long >::value_type((*iter)->_id, 0));
	  //	  _missMap.insert(map <long, long >::value_type((*iter)->_id, 0));
	  _hitMap.insert(map <Agent *, long >::value_type(*iter, 0));
	  _missMap.insert(map <Agent *, long >::value_type(*iter, 0));

	  activeAgents.push_back(*iter);

	  _meanAge[curr] += epoch - (*iter)->_birthEpoch;
	  _meanSize[curr] += (*iter)->size();
	  _meanRate[curr] += (*iter)->_profitRate;
	  _meanProfit[curr] += ((*iter)->_sumProfitReward  - (*iter)->_sumProfitDue) / (1 + (*iter)->_numProfitWon);
	  _numActive[curr]++;

	  _actionCount[_data->getCategoryIndex((*iter)->_action)]++; // of the index(_action)th category
	}
    }

  if(activeAgents.size() < 1)
    die(__FILE__, __FUNCTION__, __LINE__, "no active agents");

  _meanAge[curr] = _meanAge[curr] / _numActive[curr];
  _meanSize[curr] = _meanSize[curr] / _numActive[curr];
  _meanRate[curr] = _meanRate[curr] / _numActive[curr];
  _meanProfit[curr] = _meanProfit[curr] / _numActive[curr];

  vector < Agent * >::iterator begin = activeAgents.begin();
  vector < Agent * >::iterator end = activeAgents.end();
  
  Agent *winner;
  
  double *REG = (double *) alloca(REGISTERS * sizeof(double));
  
  for(int pat = 0; pat < _data->_data.size(); pat++) // consider each pattern
    {
      for(int i = 0; i < maxBidMatrix.size(); i++) // reset max bids
	maxBidMatrix[i] = 0;
      
      // initialize winner to first active agent

      winner = *begin;
      winner->_dummy = winner->bid(_data->feature(pat), REG);
      
      matInd = _data->getCategoryIndex(_data->category(pat)) * _data->_categories.size() + 
	_data->getCategoryIndex(winner->_action);
      
      _sumBidMatrix[matInd] += winner->_dummy; // add bid to sum
      
      if(winner->_dummy > maxBidMatrix[matInd]) // check if bid is max
	maxBidMatrix[matInd] = winner->_dummy;
      
      // look at the remaining active agents the same way
      
      iter = begin;
      for(iter++; iter != end; iter++)
	{
	  matInd = _data->getCategoryIndex(_data->category(pat)) * _data->_categories.size() + 
	    _data->getCategoryIndex((*iter)->_action);
	  
	  (*iter)->_dummy = (*iter)->bid(_data->feature(pat), REG);
	  
	  if((*iter)->outbids(*winner))
	    winner = *iter;
	  
	  _sumBidMatrix[matInd] += (*iter)->_dummy;
	  
	  if((*iter)->_dummy > maxBidMatrix[matInd])
	    maxBidMatrix[matInd] = (*iter)->_dummy;
	}

      // got winner by this point
      
      if(_data->category(pat) == winner->_action) // correct
	_hitMap[winner]++;
      else
	_missMap[winner]++;
      
      // (index of cat of pat) * (number of cats) + (assigned cat)
      _confMatrix[_data->getCategoryIndex(_data->category(pat)) * _data->_categories.size() +
		  _data->getCategoryIndex(winner->_action)]++;
      
      for(int i = 0; i < _sumMaxBidMatrix.size(); i++) // maxBidMatrix now stores the maximum bids, so add these to the total
	_sumMaxBidMatrix[i] += maxBidMatrix[i];
      
// #ifdef CCM_DEBUG
//       cout << DEBUG << " " << _data->_dataFile << " " << pat << " " << *winner;
//       cout << " " << _data->category(pat) << " " << winner->_dummy << endl;
// #endif
    }

  // calculate hits and set flags
  // if agent is not in activeAgents its _active flag is set to false already
  
  for(iter = begin; iter != end; iter++)
    {
      if(_hitMap[*iter] + _missMap[*iter] > 0) // _active is true and agent remains active
	{
	  _hits[curr] += _hitMap[*iter];
	  _numWinners[curr]++; // agent was winner at some point

	  // _active flag set to true already
	}
      else // _active true and agent not active
	{
	  // october 31, 2006	  (*iter)->_active = false;
	}
    }

  // calculate wealth percentiles OVER ENTIRE POPULATION
  
//   sort(pop._agents.begin(), pop._agents.end(), agentPtrGreaterW()); // ordered by wealth, non-increasing
  
//   _wp0 = pop[pop._agents.size() - 1]->_W; // poorest
//   _wp25 = pop[(int) (0.75 * pop._agents.size())]->_W;
//   _wp50 = pop[(int) (0.50 * pop._agents.size())]->_W;
//   _wp75 = pop[(int) (0.25 * pop._agents.size())]->_W;
//   _wp99 = pop[0]->_W; // wealthiest
}

void Stats::print(char *prefix, bool verbose)
{
  long r = numRecords();
  
  map < Agent *, long >::iterator mapIter;

  double lastHits; // running hits average
  int window = 10; // for lastHits
  int index;
  
  if(r < 1)
    die(__FILE__, __FUNCTION__, __LINE__, "no records to print");
  
  r -= 1;

  // calculate mean hits over last window updates

  for(index = r, lastHits = 0; index >= 0 && index > r - window; index--)
    lastHits += _hits[index];
  lastHits = lastHits / (r - index);

  cout << STATS;
  cout << " " << prefix;
  cout << " " << _data->_dataFile;
  cout << " E" << _epochs[r];
  cout << " " << _epochs[r];
  cout << " H " << _hits[r] << " " << (double) _hits[r] / _data->_data.size();
  //  cout << " LH " << lastHits << " " << (double) lastHits / _data->_data.size();
  cout << " AGE " << _meanAge[r];
  cout << " SIZE " << _meanSize[r];
  cout << " ACT " << _numActive[r] << " " << _numWinners[r] << "";
  //  cout << " MR " << _meanRate[r];
  //  cout << " MP " << _meanProfit[r];
  cout << " CATS";
  for(int i = 0; i < _data->_categories.size(); i++)
    cout << " " << _data->getCategory(i);
  cout << " CONF";
  for(int i = 0; i < _confMatrix.size(); i++)
    cout << " " << _confMatrix[i];
  cout << endl;
  
  if(verbose);
  //  if(verbose)
//     {
//       cout << WEALT;
//       cout << " " << prefix;
//       cout << " " << _data->_dataFile;
//       cout << " E" << _epochs[r];
//       cout << " " << _epochs[r];
//       cout << " " << _wp0 << " " << _wp25 << " " << _wp50 << " " << _wp75 << " " << _wp99 << endl;
//     }
    
  cout << PART1;
  cout << " " << prefix;
  cout << " " << _data->_dataFile;
  cout << " E" << _epochs[r];
  cout << " " << _epochs[r];
  for(mapIter = _hitMap.begin(); mapIter != _hitMap.end(); mapIter++)
    {
      cout << " L " << (mapIter->first)->id() << " " << (mapIter->first)->action() << " " << mapIter->second << " " << _missMap[mapIter->first];
    }
  cout << endl;
  
  vector < long > counts;
  cout << PART2;
  cout << " " << prefix;
  cout << " " << _data->_dataFile;
  cout << " E" << _epochs[r];
  cout << " " << _epochs[r];
  cout << " ID";
  for(mapIter = _hitMap.begin(); mapIter != _hitMap.end(); mapIter++)
    cout << " " << mapIter->first->id();
  cout << " CLASS";
  for(mapIter = _hitMap.begin(); mapIter != _hitMap.end(); mapIter++)
    cout << " " << mapIter->first->action();
  cout << " PART";
  for(mapIter = _hitMap.begin(); mapIter != _hitMap.end(); mapIter++)
    {
      cout << " " << mapIter->second + _missMap[mapIter->first];
      counts.push_back(mapIter->second + _missMap[mapIter->first]);
    }
  cout << endl;
  sort(counts.begin(), counts.end());

  cout << PART3;
  cout << " " << prefix;
  cout << " " << _data->_dataFile;
  cout << " E" << _epochs[r];
  cout << " " << _epochs[r];
  for(int i = 0; i < counts.size(); i++)
    {
      cout << " " << counts[i];
    }
  cout << endl;

  cout << BIDSM;
  cout << " " << prefix;
  cout << " " << _data->_dataFile;
  cout << " E" << _epochs[r];
  cout << " " << _epochs[r];
  for(int i = 0; i < _data->_categories.size(); i++)
    cout << " C " << _data->_categories[i] << " " <<  _actionCount[i] << "";
  cout << " MEAN";
  for(int i = 0; i < _data->_categories.size(); i++) // category
    {
      for(int j = 0; j < _data->_categories.size(); j++) // action
	{
	  // sum of all bids on patterns of ith category of agents of jth class

	  if(_data->_categoryCount[i] == 0 || _actionCount[j] == 0)
	    cout << " 0";
	  else
	    cout << " " << _sumBidMatrix[i * _data->_categories.size() + j] / _data->_categoryCount[i] / _actionCount[j];
	}
    }
  cout << " MAX";
  for(int i = 0; i < _data->_categories.size(); i++) // category
    {
      for(int j = 0; j < _data->_categories.size(); j++) // action
	{
	  // max of all bids on pattern of ith category of agents of jth class, summed over patterns of ith cat
	  // output the average max bid for each cat i and action j

	  if(_data->_categoryCount[i] == 0)
	    cout << " 0";
	  else
	    cout << " " << _sumMaxBidMatrix[i * _data->_categories.size() + j] / _data->_categoryCount[i];
	}
    }
  cout << endl;
}

