#include "args.hpp"

ostream & operator<<(ostream &os, const Args &args)
{ 
  map < string, long >::const_iterator longIter;
  map < string, double >::const_iterator doubleIter;
  map < string, string >::const_iterator stringIter;

  for(longIter = args._longArgs.begin(); longIter != args._longArgs.end(); longIter++)
    os << "long " << longIter->first << " " << longIter->second << endl;

  for(doubleIter = args._doubleArgs.begin(); doubleIter != args._doubleArgs.end(); doubleIter++)
    os << "double " << doubleIter->first << " " << doubleIter->second << endl;

  for(stringIter = args._stringArgs.begin(); stringIter != args._stringArgs.end(); stringIter++)
    os << "string " << stringIter->first << " " << stringIter->second << endl;

  return os; 
}

Args::Args(string fileName)
{
  ifstream infile(fileName.c_str(), ios::in);
  
  if(infile == NULL)
    die(__FILE__, __FUNCTION__, __LINE__, "cannot open argument file");
  
  do
    {
      string type, key;

      long longVal;
      double doubleVal;
      string stringVal;
      
      if(infile) infile >> type; else break;
      if(infile) infile >> key; else break;

      if(type == "long")
	{
	  if(infile) infile >> longVal; else break;
	  _longArgs.insert(map < string, long >::value_type(key, longVal));
	}
      else if(type == "double")
	{
	  if(infile) infile >> doubleVal; else break;
	  _doubleArgs.insert(map < string, double >::value_type(key, doubleVal));
	}
      else if(type == "string")
	{
	  if(infile) infile >> stringVal; else break;
	  _stringArgs.insert(map < string, string >::value_type(key, stringVal));
	}
      else
	{
	  die(__FILE__, __FUNCTION__, __LINE__, "cannot open argument file");
	}
      
    } while(true);

  infile.close();
}

long Args::getLong(const string key) const
{
  map < string, long >::const_iterator iter = _longArgs.find(key);

  if(iter == _longArgs.end())
     die(__FILE__, __FUNCTION__, __LINE__, "cannot get long argument");

  return iter->second;
}

double Args::getDouble(const string key) const
{
  map < string, double >::const_iterator iter = _doubleArgs.find(key);

  if(iter == _doubleArgs.end())
     die(__FILE__, __FUNCTION__, __LINE__, "cannot get double argument");

  return iter->second;
}

string Args::getString(const string key) const
{
  map < string, string >::const_iterator iter = _stringArgs.find(key);

  if(iter == _stringArgs.end())
     die(__FILE__, __FUNCTION__, __LINE__, "cannot get string argument");

  return iter->second;
}
