#include "dataset.hpp"

ostream & operator<<(ostream &os, const DataSet &data)
{
  vector < int >::const_iterator sIter;
  vector < pair < double *, int > * >::const_iterator vIter;
  
  os << *data._categories.begin();
  for(sIter = data._categories.begin() + 1; sIter != data._categories.end(); sIter++)
    os << " " << *sIter;
  os << endl;
  
  for(vIter = data._data.begin(); vIter != data._data.end(); vIter++)
    {
      os << ((*vIter)->first)[0];
      for(int j = 1; j < data._dim; j++)
	os << " " << ((*vIter)->first)[j];
      
      os << " " << (*vIter)->second << endl;
    }
  
  return os;
}

DataSet::DataSet(string dataFile, int dim, long numPatterns)
{
  set < int > categories;
  set < int >::iterator iter;

  map < int, long > categoryCount;
  
  _dataFile = dataFile;
  _dim = dim;
  
  _data.reserve(numPatterns);
  
  ifstream infile(dataFile.c_str(), ios::in);
  
  if(!infile)
    die(__FILE__, __FUNCTION__, __LINE__, "cannot open dataset");

  for(int i = 0; i < numPatterns; i++)
    {
      pair< double *, int > *p = new pair< double *, int >;
      p->first = new double[_dim];

      for(int j = 0; j < _dim; j++)
	{
	  if(!infile)
	    die(__FILE__, __FUNCTION__,  __LINE__, "bad pattern");
	  
	  infile >> p->first[j]; 
	}
      
      if(!infile)
	die(__FILE__, __FUNCTION__,  __LINE__, "bad pattern class");
      
      infile >> p->second;
      
      categories.insert(p->second);
      
      if(categoryCount.count(p->second) == 0)
	categoryCount.insert(map < int, long >::value_type(p->second, 1));
      else
	categoryCount[p->second]++;
      
      _data.push_back(p);
    }
  
  for(iter = categories.begin(); iter != categories.end(); iter++)
    {
      _categories.push_back(*iter);
    }

  // Sort categories, then create category->index map.
  
  sort(_categories.begin(), _categories.end());

  for(int i = 0; i < _categories.size(); i++)
    {
      _categoryMap.insert(map < int, int >::value_type(_categories[i], i));
      _categoryCount.push_back(categoryCount[_categories[i]]);

      cout << "dataset " << dataFile << " category at index " << i << " is " << _categories[i] ;
      cout << " number of patterns of " << _categories[i] << " is " << _categoryCount[i];
      cout << " (=" << getCategoryCount(_categories[i]) << ")" << endl;
    }
  
  infile.close();
}

DataSet::~DataSet()
{
  for(int i = 0; i  < _data.size(); i++)
    {
      delete [] _data[i]->first;
      delete _data[i];
    }
}

int DataSet::getCategory(int i)
{
  if(!(i >= 0 && i < _categories.size())) 
    die(__FILE__, __FUNCTION__, __LINE__, "bad category index"); 
  
  return _categories[i];
}

int DataSet::getCategoryIndex(int c)
{ 
  map < int, int >::iterator iter = _categoryMap.find(c); 
  
  if(iter == _categoryMap.end()) 
    die(__FILE__, __FUNCTION__, __LINE__, "bad category"); 

  return (*iter).second; 
} 

double * DataSet::feature(long pat)
{
  if(!(pat >= 0 && pat < _data.size()))
    die(__FILE__, __FUNCTION__,  __LINE__, "bad pattern index");
  
  return _data[pat]->first;
}

int DataSet::category(long pat)
{ 
  if(!(pat >= 0 && pat < _data.size()))
    die(__FILE__, __FUNCTION__,  __LINE__, "bad pattern index");
   
  return _data[pat]->second; 
}
