#ifndef POINT_H
#define POINT_H

#include <map>

class Point
{
  
public:

  friend ostream & operator<<(ostream &os, const Point &p) {
    os << p._id; map <long, double >::const_iterator iter;
    for(iter = p._outcomes.begin(); iter != p._outcomes.end(); iter++)
      os << " (" << iter->first << ", " << iter->second << ")";
    return os;
  }
  
  Point(long id, int cat) { _id = id; _category = cat; }

  inline long id() { return _id; }
  inline int category() { return _category; }
  
  inline void clearOutcome() { _outcomes.clear(); }

  inline void setOutcome(long learner, double outcome) { _outcomes[learner] = outcome; }
  
  inline double getOutcome(long learner) { 
    map < long, double > :: iterator iter = _outcomes.find(learner);
    if(iter == _outcomes.end())
      die(__FILE__, __FUNCTION__, __LINE__, "no outcome for learner");
    return iter->second;
  }

  inline bool old() { return _old; }
  inline void old(bool o) { _old = o; }
  
  inline bool active() { return _active; }
  inline void active(bool a) { _active = a; }
  
private:
  
  long _id;
  bool _old;
  int _category;
  map < long, double > _outcomes;
  bool _active;

};

class PointIdEquals:public std::unary_function < Point *, bool >
{
  
public:
  
  PointIdEquals(long id) { _id = id; }
  bool operator()(Point *p)  { return _id == p->id(); }
  
private:

  long _id;

};


struct pointPtrSmallerClass:public binary_function < Point *, Point *, bool >
{
  bool operator()(Point *lhs, Point *rhs) { return lhs->category() < rhs->category(); }
};



#endif
