#ifndef STATS_H
#define STATS_H

#include <vector>
#include <map>
#include <iostream>
#include <cfloat>

#include "dataset.hpp"
#include "agent.hpp"
#include "pop.hpp"

using namespace std;

class Stats
{
  
public:
  
  Stats(DataSet *data) { _data = data; } // stats for a dataset
  
  void update(Pop &, long); // update stats using active agents, mark inactive agents
  void print(char *, bool); // print stats

  inline long numRecords() { return _epochs.size(); } // number of times update() called

  DataSet *_data;
  
  // the following stats are stored for all the calls to update
  
  vector < long > _epochs; // value for epoch passed to each call to update()

  vector < double > _meanAge; // mean age, over all agents flagged as active
  vector < double > _meanSize; // mean size, over all agents flagged as active
  vector < double > _meanRate; // mean profit rate, over all agents flagged as active
  vector < double > _meanProfit; // mean profit, whenever realized, over all agents flagged as active
  vector < int > _numActive; // number of agents flagged as active

  vector < int > _numWinners; // number of agents that won an auction, out of those flagged as active

  vector < long > _hits; // hits

  // the following stats are stored for only the last call to update

  vector < long > _confMatrix; // confusion matrix, no. times ith cat. classified as jth at [i * no. cats + j]
  map < Agent *, long > _hitMap; // agent id -> hits
  map < Agent *, long > _missMap; // agent id -> misses

  vector < long > _actionCount; // count number of agents of each action

  vector < double > _sumBidMatrix; // no. cats * no. cats
                                   // [i * no. cats + j] stores SUM of bids on ith cat. of agents of jth cat
                                   // total contributions is no. patterns of ith cat * no. agents of jth cat
                                   // for each pattern of ith cat, consider agents of jth cat, sum their bids
  vector < double > _sumMaxBidMatrix; // no. cats * no. cats
                                      // [i * no. cats + j] stores max bid out of agents of jth cat, summed over patterns of ith cat
                                      // total contributions is no. patterns of ith cat, each contribution is a max bid for a pattern
                                      // for each pattern of ith cat, consider agents of jth cat, find max bid for that pattern, sum these max bids

  double _wp0; // wealth of agent in 0th percentile
  double _wp25;
  double _wp50;
  double _wp75;
  double _wp99; // wealth of agent in 99th percentile

};

#endif
